(*
    Persistent options management
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Config;

interface

uses
  Classes, IniFiles;

type
  TConfigItem = class
    FName: string;
  protected
    constructor Create( const AName: string );

    // Returns true if the item has a non-default value and must be
    // saved, false otherwise. The base implementation always returns true.
    function HasChanged: boolean; virtual;

  public
    property Name: string read FName;
  end;

  TConfigString = class(TConfigItem)
    FValue: string;
    FDefault: string;
  public
    constructor Create( const AName: string; const AValue: string );

    function HasChanged: boolean; override;

    property Value: string read FValue write FValue;
    property DefaultValue: string read FDefault;
  end;

  TConfigInteger = class(TConfigItem)
    FValue: integer;
    FDefault: integer;
  public
    constructor Create( const AName: string; AValue: integer );

    function HasChanged: boolean; override;

    property Value: integer read FValue write FValue;
    property DefaultValue: integer read FDefault;
  end;

  TConfigBoolean = class(TConfigItem)
    FValue: boolean;
    FDefault: boolean;
  public
    constructor Create( const AName: string; AValue: boolean );

    function HasChanged: boolean; override;

    property Value: boolean read FValue write FValue;
    property DefaultValue: boolean read FDefault;
  end;

  TConfigStringList = class(TConfigItem)
    FStrings: TStringList;

    function GetCount: integer;
  public
    constructor Create( const AName: string );
    destructor Destroy; override;

    property Strings: TStringList read FStrings;
    property Count: integer read GetCount;
  end;

  TConfigGroup = class(TConfigItem)
    FItems: TList;

    function GetItems( AIndex: integer ): TConfigItem;
    function GetCount: integer;
  public
    constructor Create( const AName: string );
    destructor Destroy; override;

    procedure Add( AItem: TConfigItem );

    property Items[ AIndex: integer ]: TConfigItem read GetItems;
    property Count: integer read GetCount;
  end;

  TConfigManager = class
    FList: TList;

    function GetCount: integer;
    function GetItem( AIndex: integer ): TConfigItem;

  protected

    constructor Create;

  public
    destructor Destroy; override;

    function Find( const AName: string ): TConfigItem;
    procedure Add( AItem: TConfigItem );
    procedure Remove( AItem: TConfigItem );

    procedure Load; virtual; abstract;
    procedure Save; virtual; abstract;

    procedure FreeItems;

    property Item[ AIndex: integer ]: TConfigItem read GetItem;
    property Count: integer read GetCount;
  end;

  TConfigIniFile = class( TConfigManager )
    FFileName: string;
    FUserName: string;

    function GetDefaultSection: string;
  public
    constructor Create( const AFileName: string );

    property DefaultSection: string read GetDefaultSection;
    property FileName: string read FFileName;

    procedure Load; override;
    procedure LoadItem( F: TMemIniFile; const S: string; E: TConfigItem );
    procedure Save; override;
    procedure SaveItem( F: TMemIniFile; const S: string; E: TConfigItem );
  end;

implementation

uses
  SysUtils, Windows, Utils;

const
  IniSectionSuffix = '.Options';

constructor TConfigItem.Create( const AName: string );
begin
  FName := AName;
end;

function TConfigItem.HasChanged: boolean;
begin
  Result := true;
end;

constructor TConfigString.Create( const AName: string; const AValue: string );
begin
  inherited Create( AName );
  FValue := AValue;
  FDefault := AValue;
end;

function TConfigString.HasChanged: boolean;
begin
  Result := FValue <> FDefault;
end;

constructor TConfigInteger.Create( const AName: string; AValue: integer );
begin
  inherited Create( AName );
  FValue := AValue;
  FDefault := AValue;
end;

function TConfigInteger.HasChanged: boolean;
begin
  Result := FValue <> FDefault;
end;

constructor TConfigBoolean.Create( const AName: string; AValue: boolean );
begin
  inherited Create( AName );
  FValue := AValue;
  FDefault := AValue;
end;

function TConfigBoolean.HasChanged: boolean;
begin
  Result := FValue <> FDefault;
end;

constructor TConfigStringList.Create( const AName: string );
begin
  inherited Create( AName );
  FStrings := TStringList.Create;
end;

destructor TConfigStringList.Destroy;
begin
  FStrings.Free;
  inherited Destroy;
end;

function TConfigStringList.GetCount: integer;
begin
  Result := FStrings.Count;
end;

function TConfigGroup.GetItems( AIndex: integer ): TConfigItem;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FItems.Count ) then
    Result := TConfigItem( FItems[AIndex] );
end;

function TConfigGroup.GetCount: integer;
begin
  Result := FItems.Count;
end;

constructor TConfigGroup.Create( const AName: string );
begin
  inherited Create( AName );
  FItems := TList.Create;
end;

destructor TConfigGroup.Destroy;
begin
  FItems.Free;
  inherited Destroy;
end;

procedure TConfigGroup.Add( AItem: TConfigItem );
begin
  FItems.Add( AItem );
end;

constructor TConfigManager.Create;
begin
  FList := TList.Create;
end;

destructor TConfigManager.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

function TConfigManager.GetCount: integer;
begin
  Result := FList.Count;
end;

function TConfigManager.GetItem( AIndex: integer ): TConfigItem;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FList.Count ) then begin
    Result := TConfigItem( FList[AIndex] );
  end;
end;

procedure TConfigManager.Add( AItem: TConfigItem );
begin
  if( AItem <> nil )and( Find( AItem.Name ) = nil ) then begin
    FList.Add( AItem );
  end;
end;

procedure TConfigManager.Remove( AItem: TConfigItem );
begin
  FList.Remove( AItem );
end;

function TConfigManager.Find( const AName: string ): TConfigItem;
var
  I: integer;
  E: TConfigItem;
begin
  Result := nil;
  for I:=1 to FList.Count do begin
    E := TConfigItem( FList[I-1] );
    if( E.Name = AName ) then begin
      Result := E;
      break;
    end;
  end;
end;

procedure TConfigManager.FreeItems;
var
  I: integer;
begin
  for I:=1 to FList.Count do begin
    TConfigItem( FList[I-1] ).Free;
  end;
  FList.Clear;
end;

constructor TConfigIniFile.Create( const AFileName: string );
var
  N: cardinal;
  B: array[1..256] of char;
begin
  inherited Create;

  FFileName := AFileName;

  FUserName := 'Default';
  N := Sizeof( B );
  if( GetUserName( @B, N ) ) then begin
    FUserName := StrPas(@B);
  end;
end;

function TConfigIniFile.GetDefaultSection: string;
begin
  Result := FUserName + IniSectionSuffix;
end;

procedure TConfigIniFile.LoadItem( F: TMemIniFile; const S: string; E: TConfigItem );
var
  J, P: integer;
begin
  if( E is TConfigString ) then begin
    TConfigString(E).Value := Trim( F.ReadString( S, E.Name, TConfigString(E).DefaultValue ) );
  end
  else if( E is TConfigInteger ) then begin
    TConfigInteger(E).Value := F.ReadInteger( S, E.Name, TConfigInteger(E).DefaultValue );
  end
  else if( E is TConfigBoolean ) then begin
    TConfigBoolean(E).Value := F.ReadInteger( S, E.Name, BoolToInteger(TConfigBoolean(E).DefaultValue) ) <> 0;
  end
  else if( E is TConfigGroup ) then begin
    for J:=1 to TConfigGroup(E).Count do begin
      LoadItem( F, S+'.'+E.Name, TConfigGroup(E).Items[J-1] );
    end;
  end
  else if( E is TConfigStringList ) then begin
    F.ReadSectionValues( S+'.'+E.Name, TConfigStringList(E).Strings );
    for J:=1 to TConfigStringList(E).Count do begin
      P := Pos( '=', TConfigStringList(E).Strings[J-1] );
      if( P > 0 ) then begin
        TConfigStringList(E).Strings[J-1] := Trim(Copy( TConfigStringList(E).Strings[J-1], P+1, 1024 ));
      end;
    end;
  end
  else begin
    // ...else unknown type!
    raise Exception.Create( 'LoadItem: unknown config item ' + E.Name );
  end;
end;

procedure TConfigIniFile.Load;
var
  F: TMemIniFile;
  I: integer;
  E: TConfigItem;
begin
  F := TMemIniFile.Create( FFileName );
  try
    for I:=1 to Count do begin
      E := Item[I-1];
      LoadItem( F, DefaultSection, E );
    end;
  finally
    F.Free;
  end
end;

procedure TConfigIniFile.SaveItem( F: TMemIniFile; const S: string; E: TConfigItem );
var
  J: integer;
begin
  if( E.HasChanged ) then begin
    if( E is TConfigString ) then begin
      F.WriteString( S, E.Name, TConfigString(E).Value );
    end
    else if( E is TConfigInteger ) then begin
      F.WriteInteger( S, E.Name, TConfigInteger(E).Value );
    end
    else if( E is TConfigBoolean ) then begin
      F.WriteInteger( S, E.Name, BoolToInteger(TConfigBoolean(E).Value) );
    end
    else if( E is TConfigGroup ) then begin
      for J:=1 to TConfigGroup(E).Count do begin
        SaveItem( F, S+'.'+E.Name, TConfigGroup(E).Items[J-1] );
      end;
    end
    else if( E is TConfigStringList ) then begin
      for J:=1 to TConfigStringList(E).Count do begin
        F.WriteString( S+'.'+E.Name, IntToStr(J), TConfigStringList(E).Strings[J-1] );
      end;
    end
    else begin
      // ...else unknown type!
      raise Exception.Create( 'SaveItem: unknown config item ' + E.Name );
    end;
  end
  else begin
    F.DeleteKey( S, E.Name );
  end;
end;

procedure TConfigIniFile.Save;
var
  F: TMemIniFile;
  I: integer;
  E: TConfigItem;
begin
  F := TMemIniFile.Create( FFileName );
  try
    for I:=1 to Count do begin
      E := Item[I-1];
      SaveItem( F, DefaultSection, E );
    end;
    F.UpdateFile;
  finally
    F.Free;
  end
end;

end.
