(*
    LCS - A study for a simple differencing program
    Syntax highlighting
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Syntax;

interface

uses
  Classes;

type
  TSyntaxTokenKind = ( tkComment, tkIdentifier, tkString, tkText );

  TSyntaxTokenItem = class
    FKind: TSyntaxTokenKind;
    FPos: integer;
    FLen: integer;
  public
    constructor Create( AKind: TSyntaxTokenKind; APos, ALen: integer );

    property Kind: TSyntaxTokenKind read FKind write FKind;
    property Pos: integer read FPos write FPos;
    property Len: integer read FLen write FLen;
  end;

  TSyntaxLineFlag = ( lfCommentBegin );

  TSyntaxLineFlags = set of TSyntaxLineFlag;

  TSyntaxTokenizer = class
    FName: string;
    FQuoteChars: string;
    FQuoteEscape: char;
    FLineComment: string;
    FCommentBegin: string;
    FCommentEnd: string;
    FIgnoreCase: boolean;
    FIdentifiers: TStringList;

    function IsKnownIdentifier( const AName: string ): boolean;
  public
    constructor Create( const AName: string );
    destructor Destroy; override;

    procedure AddIdentifiers( AIdentifiers: string );

    procedure Tokenize( ALine: string; var AFlags: TSyntaxLineFlags; ATokens: TList );

    property Name: string read FName;
    property IgnoreCase: boolean read FIgnoreCase write FIgnoreCase;
    property QuoteChars: string read FQuoteChars write FQuoteChars;
    property QuoteEscape: char read FQuoteEscape write FQuoteEscape;
    property LineComment: string read FLineComment write FLineComment;
    property CommentBegin: string read FCommentBegin write FCommentBegin;
    property CommentEnd: string read FCommentEnd write FCommentEnd;
    property Identifiers: TStringList read FIdentifiers;
  end;

  TSyntaxLine = class
    FLine: string;
    FTokens: TList;
    FFlags: TSyntaxLineFlags;
    FTokenizer: TSyntaxTokenizer;

    function GetTokenCount: integer;
    function GetTokens( AIndex: integer ): TSyntaxTokenItem;
  public
    constructor Create( ATokenizer: TSyntaxTokenizer );

    procedure Assign( ALine: string; AContextFlags: TSyntaxLineFlags );

    property Line: string read FLine;
    property Flags: TSyntaxLineFlags read FFlags;
    property Tokens[ AIndex: integer ]: TSyntaxTokenItem read GetTokens;
    property TokenCount: integer read GetTokenCount;
  end;

implementation

uses
  SysUtils;

constructor TSyntaxTokenItem.Create( AKind: TSyntaxTokenKind; APos, ALen: integer );
begin
  FKind := AKind;
  FPos := APos;
  FLen := ALen;
end;

constructor TSyntaxTokenizer.Create( const AName: string );
begin
  FName := AName;
  FQuoteEscape := #0;
  FIgnoreCase := true;
  FIdentifiers := TStringList.Create;
  FIdentifiers.Duplicates := dupIgnore;
  FIdentifiers.Sorted := true;
end;

destructor TSyntaxTokenizer.Destroy;
begin
  FIdentifiers.Free;
  inherited Destroy;
end;

procedure TSyntaxTokenizer.AddIdentifiers( AIdentifiers: string );
var
  I: integer;
  S: string;
begin
  AIdentifiers := AIdentifiers + ',';
  while( AIdentifiers <> '' ) do begin
    I := Pos( ',', AIdentifiers );
    S := Trim( Copy( AIdentifiers, 1, I-1 ) );
    if( S <> '' ) then FIdentifiers.Add( S );
    Delete( AIdentifiers, 1, I );
  end;
end;

function SubstrAt( S: PChar; const Substr: string ): boolean;
begin
  Result := (Length(Substr) > 0) and (AnsiStrLComp(S, PChar(Substr), Length(Substr)) = 0);
end;

function IsIdentifier( const S: string; Index: integer ): boolean;
begin
  Result := ( Index > 0 )and( Index <= Length(S) )and( S[Index] in ['a'..'z','A'..'Z','_'] );
end;

function TSyntaxTokenizer.IsKnownIdentifier( const AName: string ): boolean;
var
  L, H, I, R: Integer;
begin
  if( FIgnoreCase ) then begin
    Result := FIdentifiers.IndexOf( AName ) >= 0;
  end
  else begin
    // Use a binary search for the string (assumes the list is sorted)
    Result := false;
    L := 0;
    H := FIdentifiers.Count - 1;
    while( L <= H )and( not Result ) do begin
      I := (L + H) div 2;
      R := AnsiCompareStr(FIdentifiers[I], AName);
      if( R = 0 ) then
        Result := true
      else if( R < 0 ) then
        L := I + 1
      else
        H := I - 1;
    end;
  end;
end;

procedure TSyntaxTokenizer.Tokenize( ALine: string; var AFlags: TSyntaxLineFlags; ATokens: TList );
var
  P, Q: PChar;
  M, N, O: integer;
procedure AddToken( AKind: TSyntaxTokenKind; APos, ALen: integer );
  begin
    if( ALen > 0 ) then begin
      ATokens.Add( TSyntaxTokenItem.Create(AKind, APos, ALen) );
    end;
    O := APos + ALen;
    P := PChar(ALine) + O - 1;
    M := O;
  end;
procedure FlushCurrentToken;
  begin
    if( M < O ) then AddToken( tkText, M, O-M );
  end;
begin
  P := PChar(ALine);
  M := 1;
  O := 1;
  while( P^ <> #0 ) do begin
    if( lfCommentBegin in AFlags ) then begin
      // Look for end of comment
      Q := AnsiStrPos( P, PChar(FCommentEnd) );
      if( Q = nil ) then begin
        // No end of comment, so all remaining text is commented out
        AddToken( tkComment, O, Length(ALine)-O+1 );
      end
      else begin
        // End of comment found
        AFlags := AFlags - [lfCommentBegin];
        AddToken( tkComment, O, (Q-P)+Length(FCommentEnd) );
      end;
    end
    else if SubstrAt(P,FCommentBegin) then begin
      // Begin multi line comment
      FlushCurrentToken;
      AFlags := AFlags + [lfCommentBegin];
    end
    else if SubstrAt(P,FLineComment) then begin
      // End of line comment
      FlushCurrentToken;
      AddToken( tkComment, O, Length(ALine)-O+1 );
    end
    else if IsDelimiter( FQuoteChars, ALine, O ) then begin
      // This is a quoted string
      FlushCurrentToken;
      while( O < Length(ALine) ) do begin
        Inc( O );
        if( ALine[O] = FQuoteEscape )and( ALine[O+1] = ALine[M] ) then begin
          Inc( O );
          continue;
        end;
        if( ALine[O] = ALine[M] ) then begin
          break;
        end;
      end;
      AddToken( tkString, M, O-M+1 );
    end
    else if IsIdentifier( ALine, O ) then begin
      // Check if this is a reserved identifier or not
      N := O;
      while( IsIdentifier(ALine,N) ) do
        Inc( N );
      if IsKnownIdentifier( Copy(ALine,O,N-O) ) then begin
        // This is an identifier
        FlushCurrentToken;
        AddToken( tkIdentifier, O, N-O );
      end
      else begin
        // Not an identifier
        O := N;
      end;
    end
    else begin
      // Standard text
      Inc( O );
    end;
    P := PChar(ALine) + O-1;
  end;
  FlushCurrentToken;
end;

function TSyntaxLine.GetTokenCount: integer;
begin
  Result := FTokens.Count;
end;

function TSyntaxLine.GetTokens( AIndex: integer ): TSyntaxTokenItem;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FTokens.Count ) then
    Result := TSyntaxTokenItem( FTokens[AIndex] );
end;

constructor TSyntaxLine.Create( ATokenizer: TSyntaxTokenizer );
begin
  FTokenizer := ATokenizer;
  FTokens := TList.Create;
  FFlags := [];
  FLine := '';
end;

procedure TSyntaxLine.Assign( ALine: string; AContextFlags: TSyntaxLineFlags );
begin
  FLine := ALine;
  FFlags := AContextFlags;
  FTokens.Clear;
  FTokenizer.Tokenize( ALine, FFlags, FTokens );
end;

end.
